package ru.ifmo.paraschenko;

import ru.ifmo.paraschenko.objects.trees.*;
import ru.ifmo.paraschenko.objects.automata.LinearSuffixAutomaton;
import ru.ifmo.paraschenko.objects.trees.SuffixTrie;
import ru.ifmo.paraschenko.objects.IStringProcessor;
import ru.ifmo.paraschenko.algorithms.lca.SuffixAutomatonLCA;
import ru.ifmo.paraschenko.algorithms.lca.SuffixTreeLCA;

import java.util.Scanner;
import java.util.List;
import java.util.ArrayList;
import java.util.Locale;
import java.io.PrintWriter;
import java.io.File;
import java.io.FileNotFoundException;

/**
 * Description.
 *
 * @author Dmitry Paraschenko           
 */
public class ReportGenerator implements Runnable {

    private static AbstractStringProcessorStrategy[] STRATEGIES = new AbstractStringProcessorStrategy[] {
            new AbstractStringProcessorStrategy(false, "LCA ( )") {
                private long time0;
                private long time1;
                private long time2;
                private long time3;

                public IStringProcessor createInstance(String string) throws LargeInputSizeException {
                    time0 = System.currentTimeMillis();
                    LinearSuffixAutomaton auto = new LinearSuffixAutomaton(string);
                    time1 = System.currentTimeMillis();
                    SuffixAutomatonLCA lca = new SuffixAutomatonLCA(auto);
                    time2 = System.currentTimeMillis();
                    lca.init();
                    time3 = System.currentTimeMillis();
                    return auto;
                }

                public String getTime() {
                    return String.format("%.03f (%.03f + %.03f + %.03f)",
                            (time3 - time0) / 1000.0,
                            (time1 - time0) / 1000.0,
                            (time2 - time1) / 1000.0,
                            (time3 - time2) / 1000.0
                    );
                }
                
                public double time() {
                    throw new RuntimeException();
                }
            },
            new AbstractStringProcessorStrategy(false, "LCA ()") {
                private long time0;
                private long time1;
                private long time2;

                public IStringProcessor createInstance(String string) throws LargeInputSizeException {
                    time0 = System.currentTimeMillis();
                    UkkonenSuffixTree tree = new UkkonenSuffixTree(string);
                    time1 = System.currentTimeMillis();
                    new SuffixTreeLCA(tree);
                    time2 = System.currentTimeMillis();
                    return tree;
                }

                public String getTime() {
                    return String.format("%.03f (%.03f + %.03f)", (time2 - time0) / 1000.0, (time1 - time0) / 1000.0, (time2 - time1) / 1000.0);
                }

                public double time() {
                    throw new RuntimeException();
                }
            },
            new BasicStringProcessorStrategy(false, " ") {
                public IStringProcessor createInstanceImpl(String string) throws LargeInputSizeException {
                    return new LinearSuffixAutomaton(string);
                }
            },
            new BasicStringProcessorStrategy(true, "") {
                public IStringProcessor createInstanceImpl(String string) throws LargeInputSizeException {
                    return new UkkonenSuffixTree(string);
                }
            },
            new AbstractStringProcessorStrategy(true, " <br/>+<br/> ") {
                private long time0;
                private long time1;
                private long time2;
                private long time3;

                public IStringProcessor createInstance(String string) throws LargeInputSizeException {
                    time0 = System.currentTimeMillis();
                    LinearSuffixAutomaton automaton = new LinearSuffixAutomaton(string);
                    time1 = System.currentTimeMillis();
                    AutomatonSuffixTree auto = new AutomatonSuffixTree(string, automaton);
                    time2 = System.currentTimeMillis();
                    auto.init();
                    time3 = System.currentTimeMillis();
                    return auto;
                }

                public String getTime() {
                    return String.format("%.03f (%.03f + %.03f + %.03f)",
                            (time3 - time0) / 1000.0,
                            (time1 - time0) / 1000.0,
                            (time2 - time1) / 1000.0,
                            (time3 - time2) / 1000.0
                    );
                }

                public double time() {
                    throw new RuntimeException();
                }
            },
            new BasicStringProcessorStrategy(true, " ") {
                public IStringProcessor createInstanceImpl(String string) throws LargeInputSizeException {
                    return new MinimizedSuffixTree(string);
                }
            },
            new BasicStringProcessorStrategy(false, "") {
                public IStringProcessor createInstanceImpl(String string) throws LargeInputSizeException {
                    return new SuffixTrie(string);
                }
            }
    };

//    private static AbstractStringProcessorStrategy[] STRATEGIES = new AbstractStringProcessorStrategy[] {
//            new BasicStringProcessorStrategy(false, " ") {
//                public IStringProcessor createInstanceImpl(String string) throws LargeInputSizeException {
//                    return new LinearSuffixAutomaton(string);
//                }
//            },
//            new AbstractStringProcessorStrategy(true, " <br/>+<br/> ") {
//                private long time0;
//                private long time1;
//                private long time2;
//
//                public IStringProcessor createInstance(String string) throws LargeInputSizeException {
//                    time0 = System.currentTimeMillis();
//                    LinearSuffixAutomaton automaton = new LinearSuffixAutomaton(string);
//                    time1 = System.currentTimeMillis();
//                    IStringProcessor processor = new AutomatonSuffixTree(string, automaton);
//                    time2 = System.currentTimeMillis();
//                    return processor;
//                }
//
//                public String getTime() {
//                    return String.format("%.03f (%.03f + %.03f)", (time2 - time0) / 1000.0, (time1 - time0) / 1000.0, (time2 - time1) / 1000.0);
//                }
//            },
//            new BasicStringProcessorStrategy(true, "") {
//                public IStringProcessor createInstanceImpl(String string) throws LargeInputSizeException {
//                    return new UkkonenSuffixTree(string);
//                }
//            },
//            new BasicStringProcessorStrategy(true, " ") {
//                public IStringProcessor createInstanceImpl(String string) throws LargeInputSizeException {
//                    return new MinimizedSuffixTree(string);
//                }
//            },
//            new BasicStringProcessorStrategy(false, "") {
//                public IStringProcessor createInstanceImpl(String string) throws LargeInputSizeException {
//                    return new SuffixTrie(string);
//                }
//            }
//    };

    private Scanner in;
    private PrintWriter out;
    private PrintWriter html;

    private AbstractSuffixTree suffixTree;
    private int treeCount;
    private boolean equals;

    private ReportGenerator(String file) {
        try {
            in = new Scanner(new File(file + ".in"));
            out = new PrintWriter(new File(file + ".out"));
            html = new PrintWriter(new File(file + ".html"));
        } catch (FileNotFoundException e) {
            e.printStackTrace();
            System.exit(1);
        }
    }

    private boolean testStringProcessors() {
        if (!in.hasNext()) {
            return false;
        }
        html.println("<tr>");
        String string = in.next();
        printColumn(null, String.valueOf(string.length()));
        int n = in.nextInt();
        Input[] input = new Input[n];
        for (int i = 0; i < n; i++) {
            String substring = in.next();
            input[i] = new Input(string, substring);
        }
//        input = new Input[0];

        suffixTree = null;
        treeCount = 0;
        equals = true;
        for (AbstractStringProcessorStrategy strategy: STRATEGIES) {
            testStrategy(strategy, string, input);
        }
        if (treeCount <= 1) {
            printColumn(null, "&nbsp;");
        } else if (equals) {
            printColumn("green", "Passed");
        } else {
            printColumn("red", "Failed");
        }
        if (suffixTree != null) {
            printColumn(null, String.valueOf(suffixTree.getNodeCount()));
        } else {
            printColumn(null, "&nbsp;");
        }
        html.println("</tr>");
        out.println();
        html.flush();
        out.flush();
        return true;
    }

    private void printColumn(String font, String str) {
        if (font == null) {
            html.printf("<td>%s</td>\n", str);
        } else {
            html.printf("<td><font color=\"%s\">%s</font></td>\n", font, str);
        }
    }

    private void testStrategy(AbstractStringProcessorStrategy strategy, String string, Input[] input) {
        try {
            IStringProcessor processor = strategy.createInstance(string);
            if (strategy.isCheck()) {
                if (suffixTree == null) {
                    suffixTree = (AbstractSuffixTree) processor;
                }
                treeCount++;
                equals &= suffixTree.equals(processor);
            }
            out.printf("%s has been built for string of %d characters in %s seconds.\n", strategy.getName(), string.length(), strategy.getTime());
            if (input == null || testStringProcessor(processor, input)) {
                printColumn("green", strategy.getTime());
            } else {
                printColumn("red", strategy.getTime());
                System.out.printf("%s failed on some tests.\n", strategy.getName());
            }
        } catch (LargeInputSizeException e) {
            printColumn(null, "&nbsp;");
            out.printf("%s has not been built for string of %d characters.\n", strategy.getName(), string.length());
        } finally {
            html.flush();
            out.flush();
        }
    }

    private boolean testStringProcessor(IStringProcessor processor, Input[] input) {
        boolean passedAll = true;
        long time1 = System.currentTimeMillis();
        List<String> failed = new ArrayList<String>();
        for (Input in: input) {
            if (processor.isSubstring(in.string) != in.substring ||
                    processor.isSuffix(in.string) != in.suffix) {
                failed.add(in.string);
                passedAll = false;
            }
        }
        long time2 = System.currentTimeMillis();
        out.printf("It was tested on %d tests in %.3f seconds: %d passed", input.length, (time2 - time1) / 1000.0, input.length - failed.size());
        if (failed.size() > 0) {
            out.printf(", %d failed:\n", failed.size());
            int count = 0;
            for (String str: failed) {
                out.printf("%d) \'%s\'\n", count, str);
                if (++count >= 10) {
                    break;
                }
            }
        } else {
            out.printf(".\n");
        }
        return passedAll;
    }

    private static class Input {
        private String string;
        private boolean substring;
        private boolean suffix;

        Input(String string, String substring) {
            this.string = substring;
            this.substring = string.indexOf(substring) != -1;
            this.suffix = string.endsWith(substring);
        }
    }

    public void run() {
        try {
            html.println("<html>");
            html.println("<body>");
            html.println("<table border=\"1\" cellspacing=\"0\" cellpadding=\"3\">");
            html.println("<tr>");
            html.println("<th>Length</th>");
            for (AbstractStringProcessorStrategy strategy: STRATEGIES) {
                html.printf("<th>%s</th>\n", strategy.getName());
            }
            html.println("<th>Validation</th>");
            html.println("<th>Size</th>");
            html.println("</tr>");
            while (testStringProcessors()) {
            }
            html.println("</table>");
            html.println("</body>");
            html.println("</html>");
        } finally {
            out.close();
            html.close();
        }
    }

    public static void main(String[] args) {
        Locale.setDefault(Locale.US);
        for (String file: args) {
            System.out.printf("Processing file \'%s\'\n", file);
            Thread thread = new Thread(new ReportGenerator(file));
            thread.start();
        }
    }
}
