package ru.ifmo.paraschenko;

import ru.ifmo.paraschenko.objects.IStringProcessor;
import ru.ifmo.paraschenko.objects.trees.*;
import ru.ifmo.paraschenko.objects.automata.LinearSuffixAutomaton;
import ru.ifmo.paraschenko.algorithms.lca.SuffixAutomatonLCA;
import ru.ifmo.paraschenko.algorithms.lca.SuffixTreeLCA;

import java.util.*;
import java.io.PrintWriter;
import java.io.File;
import java.io.FileNotFoundException;

/**
 * Description.
 *
 * @author Dmitry Paraschenko
 */
public class StatisticsGenerator implements Runnable {

    private static double LCA_auto;
    private static double LCA_tree;

    private static double build_auto;
    private static double build_tree;

    private static final int MIN = 50;
    private static AbstractStringProcessorStrategy[] STRATEGIES = new AbstractStringProcessorStrategy[] {
            new AbstractStringProcessorStrategy(false, "LCA ( )") {
                private long time0;
                private long time1;
                private long time2;
                private long time3;

                public IStringProcessor createInstance(String string) throws LargeInputSizeException {
                    time0 = System.currentTimeMillis();
                    LinearSuffixAutomaton auto = new LinearSuffixAutomaton(string);
                    time1 = System.currentTimeMillis();
                    SuffixAutomatonLCA lca = new SuffixAutomatonLCA(auto);
                    time2 = System.currentTimeMillis();
                    lca.init();
                    time3 = System.currentTimeMillis();
                    LCA_auto = Math.max(MIN, time3 - time0);
                    return auto;
                }

                public String getTime() {
                    return String.format("%.03f (%.03f + %.03f + %.03f)",
                            (time3 - time0) / 1000.0,
                            (time1 - time0) / 1000.0,
                            (time2 - time1) / 1000.0,
                            (time3 - time2) / 1000.0
                    );
                }

                public double time() {
                    return (time3 - time0) / 1000.0; 
                }
            },
            new AbstractStringProcessorStrategy(false, "LCA ()") {
                private long time0;
                private long time1;
                private long time2;

                public IStringProcessor createInstance(String string) throws LargeInputSizeException {
                    time0 = System.currentTimeMillis();
                    UkkonenSuffixTree tree = new UkkonenSuffixTree(string);
                    time1 = System.currentTimeMillis();
                    new SuffixTreeLCA(tree);
                    time2 = System.currentTimeMillis();
                    LCA_tree = Math.max(MIN, time2 - time0);
                    return tree;
                }

                public String getTime() {
                    return String.format("%.03f (%.03f + %.03f)", (time2 - time0) / 1000.0, (time1 - time0) / 1000.0, (time2 - time1) / 1000.0);
                }

                public double time() {
                    return (time2 - time0) / 1000.0;
                }
            },
            new AbstractStringProcessorStrategy(true, "") {
                private long time0;
                private long time1;
                
                public IStringProcessor createInstance(String string) throws LargeInputSizeException {
                    time0 = System.currentTimeMillis();
                    UkkonenSuffixTree tree = new UkkonenSuffixTree(string);
                    time1 = System.currentTimeMillis();
                    build_tree = Math.max(MIN, time1 - time0);
                    return tree;
                }
                
                public String getTime() {
                    return String.format("%.03f", (time1 - time0) / 1000.0);
                }

                public double time() {
                    return (time1 - time0) / 1000.0; 
                }
            },
            new AbstractStringProcessorStrategy(true, " <br/>+<br/> ") {
                private long time0;
                private long time1;
                private long time2;
                private long time3;

                public IStringProcessor createInstance(String string) throws LargeInputSizeException {
                    time0 = System.currentTimeMillis();
                    LinearSuffixAutomaton automaton = new LinearSuffixAutomaton(string);
                    time1 = System.currentTimeMillis();
                    AutomatonSuffixTree auto = new AutomatonSuffixTree(string, automaton);
                    time2 = System.currentTimeMillis();
                    auto.init();
                    time3 = System.currentTimeMillis();
                    build_auto = Math.max(MIN, time3 - time0);
                    return auto;
                }

                public String getTime() {
                    return String.format("%.03f (%.03f + %.03f + %.03f)",
                            (time3 - time0) / 1000.0,
                            (time1 - time0) / 1000.0,
                            (time2 - time1) / 1000.0,
                            (time3 - time2) / 1000.0
                    );
                }

                public double time() {
                    return (time3 - time0) / 1000.0;
                }
            },
            new BasicStringProcessorStrategy(false, " ") {
                public IStringProcessor createInstanceImpl(String string) throws LargeInputSizeException {
                    return new LinearSuffixAutomaton(string);
                }
            },
//            new BasicStringProcessorStrategy(true, " ") {
//                public IStringProcessor createInstanceImpl(String string) throws LargeInputSizeException {
//                    return new MinimizedSuffixTree(string);
//                }
//            },
//            new BasicStringProcessorStrategy(false, "") {
//                public IStringProcessor createInstanceImpl(String string) throws LargeInputSizeException {
//                    return new SuffixTrie(string);
//                }
//            }
            new AbstractStringProcessorStrategy(true, "LCA_tree/LCA_auto") {
                public IStringProcessor createInstance(String string) throws LargeInputSizeException {
                    return null;
                }

                public String getTime() {
                    return String.format("%.03f", LCA_tree / LCA_auto);
                }

                public double time() {
                    return LCA_tree / LCA_auto;
                }
            },
            new AbstractStringProcessorStrategy(true, "build_tree/build_auto") {
                public IStringProcessor createInstance(String string) throws LargeInputSizeException {
                    return null;
                }

                public String getTime() {
                    return String.format("%.03f", build_tree / build_auto);
                }

                public double time() {
                    return build_tree / build_auto; 
                }
            },
    };

    private PrintWriter out;
    private PrintWriter html;

    private StatisticsGenerator(String file) {
        try {
            out = new PrintWriter(new File(file + ".out"));
            html = new PrintWriter(new File(file + ".html"));
        } catch (FileNotFoundException e) {
            e.printStackTrace();
            System.exit(1);
        }
    }

    private boolean testStringProcessors() {
        testStringProcessors(2, new String[] {"a", "b"});

//        for (int len = 1; len <= 10; len++) {
//            for (int size = 500; size <= 5000; size += 500) {
//                testStringProcessors(2, generateLang(2, len, size));
//            }
//        }
//        for (int lang = 1; lang < 10000; lang *= 2) {
//            for (int len = 1; len < 10000; len *= 2) {
//                testStringProcessors(26, generateLang(26, lang, len));
//            }
//        }
        return false;
    }

    private void testStringProcessors(int alpha, String[] lang) {
//        int len = 10000;
//        for (int i = 4; i <= 6; i++) {
//            testStringProcessors(alpha, lang, len);
//            len *= 10;
//        }
        for (int len = 100000; len <= 1000000; len += 100000) {
            testStringProcessors(alpha, lang, len);
        }
    }

    private String[] generateLang(int alpha, int length, int size) {
        String[] lang = new String[size];
        for (int i = 0; i < size; i++) {
            lang[i] = generateWord(alpha, length);
        }
        return lang;
    }

    private String generateWord(int alpha, int length) {
        StringBuilder sb = new StringBuilder();
        for (int i = 0; i < length; i++) {
            sb.append(generateCharacter(alpha));
        }
        return sb.toString();
    }

    private char generateCharacter(int alpha) {
        int c = random.nextInt(alpha);
        if (c < 26) {
            return (char) ('a' + c);
        } else if (c < 52) {
            return (char) ('A' + c - 26);
        } else if (c < 62) {
            return (char) ('0' + c - 52);
        } else {
            return 0;
        }
    }

    Random random = new Random();

    private boolean testStringProcessors(int alpha, String[] lang, int length) {
        System.out.printf("alpha: %d   lang: %d    length: %d", alpha, lang.length, length);
        double[] times = new double[STRATEGIES.length];
        final int count = 10;
        for (int i = 0; i < count; i++) {
            System.out.print("|");
            StringBuilder sb = new StringBuilder();
            while (sb.length() < length) {
                sb.append(lang[random.nextInt(lang.length)]);
            }
            String string = sb.toString();
            for (int ii = 0; ii < STRATEGIES.length; ii++) {
                AbstractStringProcessorStrategy strategy = STRATEGIES[ii];
//                testStrategy(strategy, string);
                try {
                    strategy.createInstance(string);
                    times[ii] += strategy.time();
                } catch (LargeInputSizeException e) {
                }
                System.out.print(".");
            }
        }
        System.out.println("|");
        html.println("<tr>");
        printColumn(null, "" + alpha);
        printColumn(null, lang.length + " x " + lang[0].length());
        printColumn(null, String.valueOf(length));
        for (double x: times) {
            printColumn(null, String.format("%.03f", x / count));
        }
        html.println("</tr>");
        html.flush();
        return false;
    }

    private void printColumn(String font, String str) {
        if (font == null) {
            html.printf("<td>%s</td>\n", str);
        } else {
            html.printf("<td><font color=\"%s\">%s</font></td>\n", font, str);
        }
    }

    private void testStrategy(AbstractStringProcessorStrategy strategy, String string) {
        try {
            IStringProcessor proc = strategy.createInstance(string);
            String time = strategy.getTime();
            printColumn(proc != null || time.startsWith("0.") || time.startsWith("1.00") ? "green" : "red", time);
        } catch (LargeInputSizeException e) {
            printColumn(null, "&nbsp;");
            out.printf("%s has not been built for string of %d characters.\n", strategy.getName(), string.length());
        } finally {
            html.flush();
            out.flush();
        }
    }

    public void run() {
        try {
            html.println("<html>");
            html.println("<body>");
            html.println("<table border=\"1\" cellspacing=\"0\" cellpadding=\"3\">");
            html.println("<tr>");
            html.println("<th>Alphabet</th>");
            html.println("<th>Language</th>");
            html.println("<th>Length</th>");
            for (AbstractStringProcessorStrategy strategy: STRATEGIES) {
                html.printf("<th>%s</th>\n", strategy.getName());
            }
            html.println("</tr>");
            while (testStringProcessors()) {
            }
            html.println("</table>");
            html.println("</body>");
            html.println("</html>");
        } finally {
            out.close();
            html.close();
        }
    }

    public static void main(String[] args) {
        Locale.setDefault(Locale.US);
        new Thread(new StatisticsGenerator("statistics")).start();
    }
}
