package ru.ifmo.paraschenko.objects.automata;

import ru.ifmo.paraschenko.objects.graphs.AdjacencyListGraph;
import ru.ifmo.paraschenko.objects.ISuffixAutomaton;
import ru.ifmo.paraschenko.LargeInputSizeException;

public final class LinearSuffixAutomaton extends AdjacencyListGraph implements ISuffixAutomaton {
    private int nodeCount;
    private final int[] suffix;
	private final int[] ln;
    private final boolean[] finish;
	private int lastNode;
	private int length;
	
	public LinearSuffixAutomaton(int verCnt, int edgeCnt) throws LargeInputSizeException {
        super(verCnt, edgeCnt);
        suffix = new int[verCnt];
		ln = new int[verCnt];
        finish = new boolean[verCnt];
        lastNode = createNode(0);
		length = 0;
	}
	
	public LinearSuffixAutomaton(String string) throws LargeInputSizeException {
		this(1 + 2 * string.length(), 3 * string.length());
        if (string.length() > 5000000) {
            throw new LargeInputSizeException();
        }
//        System.out.println(string);
        for (char ch: string.toCharArray()) {
			addCharacter(ch);
		}
        for (int p = lastNode; p >= 0; p = suffix[p]) {
            finish[p] = true;
		}
    }

    public int getNodeCount() {
        return nodeCount;
    }

    private int createNode(int log) {
        suffix[nodeCount] = -1;
        ln[nodeCount] = log;
		return nodeCount++;
	}

	private void addCharacter(char ch) {
		length++;
		int p = lastNode;
		lastNode = createNode(ln[p] + 1);
		while (p >= 0 && getEdge(p, ch) == -1) {
			addEdge(p, ch, lastNode);
			p = suffix[p];
		}
		if (p < 0) {
			suffix[lastNode] = 0;
			return;
		}
		int q = getTarget(getEdge(p, ch));
		if (ln[p] + 1 == ln[q]) {
			suffix[lastNode] = q;
		} else {
			int r = createNode(ln[p] + 1);
            for (int edge = first[q]; edge != -1; edge = next[edge]) {
                addEdge(r, chr[edge], getTarget(edge));
            }
			suffix[r] = suffix[q];
			suffix[q] = r;
			for (int edge = getEdge(p, ch); p >= 0 && edge >= 0 && getTarget(edge) == q; p = suffix[p], edge = p < 0 ? - 1 : getEdge(p, ch)) {
                setTarget(edge, r);
			}
			suffix[lastNode] = r;
		}
	}
	
    public boolean isFinish(int state) {
        return state >= 0 && finish[state];
    }

    public int getSuffix(int state) {
        return suffix[state];
    }

	public int length() {
		return length;
	}

    public int process(String string) {
		int state = 0;
		for (char ch: string.toCharArray()) {
			int edge = getEdge(state, ch);
			if (edge == -1) {
				return -1;
			}
            state = getTarget(edge);
        }
		return state;
	}

	public boolean isSubstring(String string) {
		return process(string) != -1;
	}

	public boolean isSuffix(String string) {
        return isFinish(process(string));
	}

	@Override
	public String toString() {
		StringBuilder sb = new StringBuilder();
		for (int i = 0; i < getNodeCount(); i++) {
			sb.append("ver: ").append(i).append(isFinish(i) ? " +" : " -").append("   ln: ").append(ln[i]).append("  suf: ").append(suffix[i]).append('\n');
            for (int edge = first[i]; edge != -1; edge = next[edge]) {
				sb.append("  ").append(chr[edge]).append(" -> ").append(getTarget(edge)).append('\n');
			}
		}
		return sb.toString();
	}


    public int getLastNode() {
        return lastNode;
    }
}
