package ru.ifmo.paraschenko.objects.trees;

import ru.ifmo.paraschenko.objects.graphs.AdjacencyListGraph;
import ru.ifmo.paraschenko.objects.ISuffixTree;
import ru.ifmo.paraschenko.LargeInputSizeException;

import java.util.Arrays;

public abstract class AbstractSuffixTree extends AdjacencyListGraph implements ISuffixTree {
    private char[] string;

	private boolean[] finish;
    private final int[] begin;
    private final int[] end;

    private int[] suffix;
    private int[] source;
    private int[] parentEdge;

    protected AbstractSuffixTree(String string, int verCnt, int edgeCnt) throws LargeInputSizeException {
        super(verCnt, edgeCnt);
        this.string = string.toCharArray();
        finish = new boolean[verCnt];
        begin = new int[edgeCnt];
        end = new int[edgeCnt];
        suffix = new int[verCnt];
        Arrays.fill(suffix, -1);
        source = new int[edgeCnt];
        Arrays.fill(source, -1);
        parentEdge = new int[verCnt];
        Arrays.fill(parentEdge, -1);
        createNode();
        suffix[0] = 0;
    }
	
    protected AbstractSuffixTree(String string, int verCnt) throws LargeInputSizeException {
        this(string, verCnt, verCnt);
    }

	public char charAt(int pos) {
		return string[pos];
	}

    public int getBegin(int edge) {
        return begin[edge];
    }

    public void setBegin(int edge, int value) {
        begin[edge] = value;
    }

    public int getEnd(int edge) {
        return end[edge];
    }

    public void setEnd(int edge, int value) {
        end[edge] = value;
    }

    public boolean isFinish(int node) {
        return finish[node];
    }

    public void setFinish(int node, boolean value) {
        finish[node] = value;
    }

    @Override
    public int createNode() {
        int node = super.createNode();
        finish[node] = false;
        return node;
    }

    public int length() {
        return string.length;
    }

    protected int addEdge(int src, int trg, int b, int e) {
        int edge = super.addEdge(src, string[b], trg);
        begin[edge] = b;
        end[edge] = e;
        source[edge] = src;
        parentEdge[trg] = edge;
        return edge;
    }

    protected int splitEdge(int edge, int len) {
        if (len == 0) {
            return - 1;
        }
        int node = createNode();
        int trg = getTarget(edge);
        addEdge(node, trg, begin[edge] + len, end[edge]);
        setTarget(edge, node);
        end[edge] = begin[edge] + len;
        parentEdge[node] = edge;
        return node;
    }

	@Override
	public String toString() {
		StringBuilder sb = new StringBuilder();
		for (int i = 0; i < getNodeCount(); i++) {
			sb.append("ver: ").append(i).append(finish[i] ? " +" : " -").append('\n');
			for (int edge = first[i]; edge != -1; edge = next[edge]) {
                sb.append("  ");
                for (int j = begin[edge]; j < end[edge]; j++) {
                    sb.append(string[j]);
                }
                sb.append(" -> ").append(getTarget(edge)).append('\n');
            }
		}
		return sb.toString();
	}

    @Override
    public boolean equals(Object o) {
        if (o instanceof AbstractSuffixTree) {
            AbstractSuffixTree that = (AbstractSuffixTree) o;
            return this.getNodeCount() == that.getNodeCount() &&
                    this.getEdgeCount() == that.getEdgeCount() &&
                    equals(this, 0, that, 0);
        }
        return false;
    }

    private static boolean equals(AbstractSuffixTree tree1, int node1, AbstractSuffixTree tree2, int node2) {
        if (node1 < 0 && node2 < 0) {
            return true;
        }
        if (node1 < 0 || node2 < 0) {
            return false;
        }
        if (tree1.finish[node1] != tree2.finish[node2]) {
            return false;
        }
        int count1 = 0;
        for (int edge1 = tree1.first[node1]; edge1 != -1; edge1 = tree1.next[edge1]) {
            int edge2 = tree2.getEdge(node2, tree1.chr[edge1]);
            if (edge2 == -1 || !equals(tree1, tree1.getTarget(edge1), tree2, tree2.getTarget(edge2))) {
                return false;
            }
            count1++;
        }
        int count2 = 0;
        for (int edge2 = tree2.first[node2]; edge2 != -1; edge2 = tree2.next[edge2]) {
            count2++;
        }
        return count1 == count2;
    }

    public int process(String string) {
        char[] str = string.toCharArray();
        int node = 0;
        for (int i = 0; i < str.length; ) {
            int edge = getEdge(node, str[i]);
            if (edge == -1) {
                return -1;
            }
            for (int j = getBegin(edge); j < getEnd(edge); j++, i++) {
                if (i == str.length) {
                    return -2;
                }
                if (charAt(j) != str[i]) {
                    return -1;
                }
            }
            node = getTarget(edge);
        }
        return node;
    }

    public boolean isSubstring(String string) {
        return process(string) != -1;
    }

    public boolean isSuffix(String string) {
        int node = process(string);
        return node >= 0 && isFinish(node);
    }

    public int getSuffix(int node) {
        if (suffix[node] != -1) {
            return suffix[node];
        }
        if (node == 0) {
            return 0;
        }
        int edge = parentEdge[node];
        int b = getBegin(edge);
        int e = getEnd(edge);
        int par = source[edge];
        if (par == 0) {
            b++;
        } else {
            par = getSuffix(par);
        }
        while (b < e) {
            edge = getEdge(par, charAt(b));
            int length = getEnd(edge) - getBegin(edge);
            if (e - b >= length) {
                b += length;
                par = getTarget(edge);
            } else {
                return -1;
            }
        }
        return suffix[node] = par;
    }

    public int getSource(int node) {
        return source[node];
    }

    public int getParentEdge(int node) {
        return parentEdge[node];
    }

    public void setSuffix(int node, int value) {
        suffix[node] = value;
    }

    public void setSource(int node, int value) {
        source[node] = value;
    }

    public void setParentEdge(int node, int value) {
        parentEdge[node] = value;
    }

    public String getString() {
        return String.valueOf(string);
    }
}

